# DifferentialDynamicsModels.jl

[![Build Status](https://travis-ci.org/schmrlng/DifferentialDynamicsModels.jl.svg?branch=master)](https://travis-ci.org/schmrlng/DifferentialDynamicsModels.jl)
[![Build status](https://ci.appveyor.com/api/projects/status/ppvx7g8cpt04fuvy?svg=true)](https://ci.appveyor.com/project/schmrlng/differentialdynamicsmodels-jl)
[![codecov.io](http://codecov.io/github/schmrlng/DifferentialDynamicsModels.jl/coverage.svg?branch=master)](http://codecov.io/github/schmrlng/DifferentialDynamicsModels.jl?branch=master)

Originally written to support [MotionPlanning.jl](https://github.com/schmrlng/MotionPlanning.jl), this package contains a core set of abstractions for working with dynamical systems of the form ![differential dynamics model](https://latex.codecogs.com/png.latex?%5Cinline%20%5Cdot%7Bx%7D%28t%29%20%3D%20f%28x%28t%29%2C%20u%28t%29%29), where ![x](https://latex.codecogs.com/png.latex?%5Cinline%20x) and ![u](https://latex.codecogs.com/png.latex?%5Cinline%20u) represent the state and control input of the system, respectively. This package is not intended to be a substitute for more comprehensive/domain-specific packages (e.g., [DynamicalSystems.jl](https://github.com/JuliaDynamics/DynamicalSystems.jl), [RigidBodyDynamics.jl](https://github.com/JuliaRobotics/RigidBodyDynamics.jl), or [DifferentialEquations.jl](https://github.com/JuliaDiffEq/DifferentialEquations.jl)); instead it aims to be as lightweight as possible while offering an extensible framework for operations on state/control trajectories relevant to robot trajectory planning.

DifferentialDynamicsModels.jl exports the following types:

- `DifferentialDynamics` — abstract type representing the ![f](https://latex.codecogs.com/png.latex?%5Cinline%20f) in ![differential dynamics model](https://latex.codecogs.com/png.latex?%5Cinline%20%5Cdot%7Bx%7D%20%3D%20f%28x%2Cu%29). Instances of concrete subtypes should be [callable](https://docs.julialang.org/en/latest/manual/methods/#Function-like-objects-1), i.e., implement `(::MyDynamics)(x, u)`. Examples include:
    - [`SingleIntegratorDynamics{N}`](https://github.com/schmrlng/DifferentialDynamicsModels.jl/blob/5111999292477f159eb444a2908ad85287185256/src/DifferentialDynamicsModels.jl#L150) defined in this package. Implements the dynamics ![single integrator dynamics](https://latex.codecogs.com/png.latex?%5Cinline%20f%28x%2C%20u%29%20%3D%20u). The type parameter `N` denotes the state and control dimension.
    - [`LinearDynamics{Dx,Du}`](https://github.com/schmrlng/LinearDynamicsModels.jl/blob/4f63c2be8c57b59a5c46f243ed0ad47213eac4de/src/LinearDynamicsModels.jl#L19) defined in [LinearDynamicsModels.jl](https://github.com/schmrlng/LinearDynamicsModels.jl). Implements the dynamics ![linear dynamics](https://latex.codecogs.com/png.latex?%5Cinline%20f%28x%2C%20u%29%20%3D%20Ax%20&plus;%20Bu%20&plus;%20c). The type parameters `Dx` and `Du` denote the state and control dimension respectively.
    - [`SimpleCarDynamics{Dv,Dκ}`](https://github.com/schmrlng/SimpleCarModels.jl/blob/8d9a97467b7b3c01733c71b46e829d0b25e6acb6/src/models.jl#L9) defined in [SimpleCarModels.jl](https://github.com/schmrlng/SimpleCarModels.jl). Implements unicycle dynamics extended by `Dv` and `Dκ` integrators in the speed and curvature controls respectively.
    - [`BicycleModel`](https://github.com/StanfordASL/Pigeon.jl/blob/48c479a801b0bc6bf11b59d312922e228d9e3481/src/vehicle_dynamics.jl#L90) defined in [Pigeon.jl](https://github.com/StanfordASL/Pigeon.jl). Implements a single-track (i.e., "bicycle") vehicle model for use in QP-based model predictive control.
- `ControlInterval` — abstract type representing a control function defined over a time interval. Concrete subtypes include:
    - [`StepControl`](https://github.com/schmrlng/DifferentialDynamicsModels.jl/blob/5111999292477f159eb444a2908ad85287185256/src/DifferentialDynamicsModels.jl#L67) (a.k.a. `ZeroOrderHoldControl`) — constant control input over a time interval.
    - [`RampControl`](https://github.com/schmrlng/DifferentialDynamicsModels.jl/blob/5111999292477f159eb444a2908ad85287185256/src/DifferentialDynamicsModels.jl#L86) (a.k.a. `FirstOrderHoldControl`) — linear interpolation between two control inputs over a time interval.
    - [`BVPControl`](https://github.com/schmrlng/DifferentialDynamicsModels.jl/blob/5111999292477f159eb444a2908ad85287185256/src/DifferentialDynamicsModels.jl#L107) — a wrapper around state/control trajectory functions produced, e.g., by an external boundary value problem solver.
- `CostFunctional` - abstract type representing an objective for optimal control. Concrete subtypes include:
    - [`Time`](https://github.com/schmrlng/DifferentialDynamicsModels.jl/blob/5111999292477f159eb444a2908ad85287185256/src/DifferentialDynamicsModels.jl#L25) — trajectory duration ![T](https://latex.codecogs.com/png.latex?%5Cinline%20T).
    - [`TimePlusQuadraticControl`](https://github.com/schmrlng/DifferentialDynamicsModels.jl/blob/5111999292477f159eb444a2908ad85287185256/src/DifferentialDynamicsModels.jl#L26) — mixed time/control effort penalty ![time plus quadratic control](https://latex.codecogs.com/svg.latex?%5Cinline%20%5Cint_0%5ET%20%281%20&plus;%20u%28t%29%5ET%20R%20u%28t%29%29%20%5Cmathop%7B%7D%5C%21%5Cmathrm%7Bd%7Dt).
- `SteeringBVP{D<:DifferentialDynamics,C<:CostFunctional,SC<:SteeringConstraints,SD<:SteeringCache}` — concrete type that serves as a container that both defines a two-point boundary value problem (i.e., "steering problem" connecting two states in robot motion planning parlance) and provides the means to solve it. Instances should be callable, i.e., users should define `(::SteeringBVP{MyDynamics,MyCostFunctional,MySteeringConstraints,MySteeringCache})` returning a named tuple `(cost=..., controls=...)`, leveraging Julia's multiple-dispatch on the `SteeringBVP` type parameters to select the right implementation. Examples include:
    - [`GeometricSteering`](https://github.com/schmrlng/DifferentialDynamicsModels.jl/blob/5111999292477f159eb444a2908ad85287185256/src/DifferentialDynamicsModels.jl#L160) — `SingleIntegratorDynamics` with a `Time` optimization objective and `BoundedControlNorm` control constraint (equivalent to optimizing Euclidean arc length).
    - [`LinearQuadraticSteering`](https://github.com/schmrlng/LinearDynamicsModels.jl/blob/master/src/LinearDynamicsModels.jl#L63) — `LinearDynamics` with a `TimePlusQuadraticControl` optimization objective and no control constraints.
    - [`DubinsSteering` and `ReedsSheppSteering`](https://github.com/schmrlng/SimpleCarModels.jl) — minimum arc length [Dubins](http://planning.cs.uiuc.edu/node821.html) and [Reeds-Shepp](http://planning.cs.uiuc.edu/node822.html) steering for a simple car with minimum turning radius.

DifferentialDynamicsModels.jl defines the following functions that can and should be extended by dependent packages if applicable:

- `state_dim(::DifferentialDynamics)` — dimension of the state ![x](https://latex.codecogs.com/png.latex?%5Cinline%20x).
- `control_dim(::DifferentialDynamics)` — dimension of the control input ![u](https://latex.codecogs.com/png.latex?%5Cinline%20u).
- `duration(controls)` — length of the time interval over which `controls` is defined; `controls` may be a single `ControlInterval` or an iterable container (e.g., a `Vector`) of `ControlInterval`s.
- `propagate(f::DifferentialDynamics, x::State, controls, ts)` — propagates the dynamics ODE `f` starting from `x` to return a state or sequence of states at times `ts`. `controls` may be a single `ControlInterval` or an iterable of `ControlInterval`s. `ts` may be a single `Number`, an iterable of `Number`s, or not be included as an argument at all, in which case it defaults to `duration(controls)`.
- `waypoints(f::DifferentialDynamics, x::State, controls, dt_or_N)` — similar to `propagate` but returns equally spaced states (in time) along the controlled trajectory. `dt_or_N` may be an `AbstractFloat` or an `Int`, corresponding to a desired spacing `dt` or a desired total number of waypoints `N`.
- `instantaneous_control(controls, ts)` — similar to `propagate` but returns the control input instead of the state at `ts`.

Performance-oriented users may note that the last three functions allocate and return `Vector`s in the case that `ts` is an iterable; this package also defines the `Propagate` and `InstantaneousControl` [iterators](https://docs.julialang.org/en/latest/manual/interfaces/#man-interface-iteration-1) constructed in the same ways that their function counterparts are called (`waypoints_itr` also exists, and returns a `Propagate` iterator).
